import json
from hashlib import sha1
from typing import Any, Optional, TypeVar, Union
from urllib.parse import urljoin

import attrs
from w3lib.encoding import (
    html_body_declared_encoding,
    html_to_unicode,
    http_content_type_encoding,
    read_bom,
    resolve_encoding,
)
from w3lib.url import canonicalize_url

from web_poet._base import _HttpHeaders
from web_poet.mixins import SelectableMixin, UrlShortcutsMixin
from web_poet.utils import _create_deprecated_class, memoizemethod_noargs

from .url import RequestUrl as _RequestUrl
from .url import ResponseUrl as _ResponseUrl

T_headers = TypeVar("T_headers", bound=_HttpHeaders)


RequestUrl = _create_deprecated_class("RequestUrl", _RequestUrl)
ResponseUrl = _create_deprecated_class("ResponseUrl", _ResponseUrl)


class HttpRequestBody(bytes):
    """A container for holding the raw HTTP request body in bytes format."""

    pass


class HttpResponseBody(bytes):
    """A container for holding the raw HTTP response body in bytes format."""

    def bom_encoding(self) -> Optional[str]:
        """Returns the encoding from the byte order mark if present."""
        return read_bom(self)[0]

    def declared_encoding(self) -> Optional[str]:
        """Return the encoding specified in meta tags in the html body,
        or ``None`` if no suitable encoding was found"""
        return html_body_declared_encoding(self)

    def json(self) -> Any:
        """
        Deserialize a JSON document to a Python object.
        """
        return json.loads(self)


class HttpRequestHeaders(_HttpHeaders):
    """A container for holding the HTTP request headers.

    It's able to accept instantiation via an Iterable of Tuples:

    >>> pairs = [("Content-Encoding", "gzip"), ("content-length", "648")]
    >>> HttpRequestHeaders(pairs)
    <HttpRequestHeaders('Content-Encoding': 'gzip', 'content-length': '648')>

    It's also accepts a mapping of key-value pairs as well:

    >>> pairs = {"Content-Encoding": "gzip", "content-length": "648"}
    >>> headers = HttpRequestHeaders(pairs)
    >>> headers
    <HttpRequestHeaders('Content-Encoding': 'gzip', 'content-length': '648')>

    Note that this also supports case insensitive header-key lookups:

    >>> headers.get("content-encoding")
    'gzip'
    >>> headers.get("Content-Length")
    '648'

    These are just a few of the functionalities it inherits from
    :class:`multidict.CIMultiDict`. For more info on its other features, read
    the API spec of :class:`multidict.CIMultiDict`.
    """

    pass


class HttpResponseHeaders(_HttpHeaders):
    """A container for holding the HTTP response headers.

    It's able to accept instantiation via an Iterable of Tuples:

    >>> pairs = [("Content-Encoding", "gzip"), ("content-length", "648")]
    >>> HttpResponseHeaders(pairs)
    <HttpResponseHeaders('Content-Encoding': 'gzip', 'content-length': '648')>

    It's also accepts a mapping of key-value pairs as well:

    >>> pairs = {"Content-Encoding": "gzip", "content-length": "648"}
    >>> headers = HttpResponseHeaders(pairs)
    >>> headers
    <HttpResponseHeaders('Content-Encoding': 'gzip', 'content-length': '648')>

    Note that this also supports case insensitive header-key lookups:

    >>> headers.get("content-encoding")
    'gzip'
    >>> headers.get("Content-Length")
    '648'

    These are just a few of the functionalities it inherits from
    :class:`multidict.CIMultiDict`. For more info on its other features, read
    the API spec of :class:`multidict.CIMultiDict`.
    """

    def declared_encoding(self) -> Optional[str]:
        """Return encoding detected from the Content-Type header, or None
        if encoding is not found"""
        content_type = self.get("Content-Type", "")
        return http_content_type_encoding(content_type)


@attrs.define(auto_attribs=False, slots=False, eq=False)
class HttpRequest:
    """Represents a generic HTTP request used by other functionalities in
    **web-poet** like :class:`~.HttpClient`.
    """

    url: _RequestUrl = attrs.field(converter=_RequestUrl)
    method: str = attrs.field(default="GET", kw_only=True)
    headers: HttpRequestHeaders = attrs.field(
        factory=HttpRequestHeaders, converter=HttpRequestHeaders, kw_only=True
    )
    body: HttpRequestBody = attrs.field(
        factory=HttpRequestBody, converter=HttpRequestBody, kw_only=True
    )

    def urljoin(self, url: Union[str, _RequestUrl, _ResponseUrl]) -> _RequestUrl:
        """Return *url* as an absolute URL.

        If *url* is relative, it is made absolute relative to :attr:`url`."""
        return _RequestUrl(urljoin(str(self.url), str(url)))


@attrs.define(auto_attribs=False, slots=False, eq=False)
class HttpResponse(SelectableMixin, UrlShortcutsMixin):
    """A container for the contents of a response, downloaded directly using an
    HTTP client.

    ``url`` should be a URL of the response (after all redirects),
    not a URL of the request, if possible.

    ``body`` contains the raw HTTP response body.

    The following are optional since it would depend on the source of the
    ``HttpResponse`` if these are available or not. For example, the responses
    could simply come off from a local HTML file which doesn't contain ``headers``
    and ``status``.

    ``status`` should represent the int status code of the HTTP response.

    ``headers`` should contain the HTTP response headers.

    ``encoding`` encoding of the response. If None (default), encoding
    is auto-detected from headers and body content.
    """

    url: _ResponseUrl = attrs.field(converter=_ResponseUrl)
    body: HttpResponseBody = attrs.field(converter=HttpResponseBody)
    status: Optional[int] = attrs.field(default=None, kw_only=True)
    headers: HttpResponseHeaders = attrs.field(
        factory=HttpResponseHeaders, converter=HttpResponseHeaders, kw_only=True
    )
    _encoding: Optional[str] = attrs.field(default=None, kw_only=True)

    _DEFAULT_ENCODING = "ascii"
    _cached_text: Optional[str] = None

    @property
    def text(self) -> str:
        """
        Content of the HTTP body, converted to unicode
        using the detected encoding of the response, according
        to the web browser rules (respecting Content-Type header, etc.)
        """
        # Access self.encoding before self._cached_text, because
        # there is a chance self._cached_text would be already populated
        # while detecting the encoding
        encoding = self.encoding
        if self._cached_text is None:
            fake_content_type_header = f"charset={encoding}"
            encoding, text = html_to_unicode(fake_content_type_header, self.body)
            self._cached_text = text
        return self._cached_text

    def _selector_input(self) -> str:
        return self.text

    @property
    def encoding(self) -> Optional[str]:
        """Encoding of the response"""
        return (
            self._encoding
            or self._body_bom_encoding()
            or self._headers_declared_encoding()
            or self._body_declared_encoding()
            or self._body_inferred_encoding()
        )

    @memoizemethod_noargs
    def json(self) -> Any:
        """Deserialize a JSON document to a Python object."""
        return self.body.json()

    @memoizemethod_noargs
    def _body_bom_encoding(self) -> Optional[str]:
        return self.body.bom_encoding()

    @memoizemethod_noargs
    def _headers_declared_encoding(self) -> Optional[str]:
        return self.headers.declared_encoding()

    @memoizemethod_noargs
    def _body_declared_encoding(self) -> Optional[str]:
        return self.body.declared_encoding()

    @memoizemethod_noargs
    def _body_inferred_encoding(self) -> Optional[str]:
        content_type = self.headers.get("Content-Type", "")
        body_encoding, text = html_to_unicode(
            content_type,
            self.body,
            # FIXME: type ignore can be removed when the following is released:
            # https://github.com/scrapy/w3lib/pull/190
            auto_detect_fun=self._auto_detect_fun,  # type: ignore[arg-type]
            default_encoding=self._DEFAULT_ENCODING,
        )
        self._cached_text = text
        return body_encoding

    def _auto_detect_fun(self, body: bytes) -> Optional[str]:
        for enc in (self._DEFAULT_ENCODING, "utf-8", "cp1252"):
            try:
                body.decode(enc)
            except UnicodeError:
                continue
            return resolve_encoding(enc)


def request_fingerprint(req: HttpRequest) -> str:
    """Return the fingerprint of the request."""
    fp = sha1()
    fp.update(req.method.encode() + b"\n")
    fp.update(canonicalize_url(str(req.url)).encode() + b"\n")
    for name, value in sorted(req.headers.items()):
        fp.update(f"{name.title()}:{value}\n".encode())
    fp.update(b"\n")
    fp.update(req.body)
    return fp.hexdigest()



from .browser import BrowserHtml, BrowserResponse
from .client import HttpClient
from .http import (
    HttpRequest,
    HttpRequestBody,
    HttpRequestHeaders,
    HttpResponse,
    HttpResponseBody,
    HttpResponseHeaders,
)
from .page_params import PageParams
from .response import AnyResponse
from .stats import Stats
from .url import RequestUrl, ResponseUrl



import asyncio
import random
from typing import Callable, List

import attrs
import pytest

from tests.po_lib_to_return import (
    CustomProductPage,
    CustomProductPageDataTypeOnly,
    CustomProductPageNoReturns,
    ImprovedProductPage,
    LessProductPage,
    MoreProductPage,
    Product,
    ProductFewerFields,
    ProductMoreFields,
    ProductPage,
    ProductSimilar,
    SimilarProductPage,
)
from web_poet import (
    HttpResponse,
    ItemPage,
    WebPage,
    field,
    item_from_fields,
    item_from_fields_sync,
)
from web_poet.fields import FieldsMixin, get_fields_dict


@attrs.define
class Item:
    name: str
    price: str


@attrs.define
class Page(ItemPage[Item]):
    response: HttpResponse

    @field
    def name(self):
        return self.response.css("title ::text").get()

    @field
    async def price(self):
        await asyncio.sleep(0.01)
        return "$123"


@attrs.define
class InvalidPage(ItemPage[Item]):
    response: HttpResponse

    @field
    def name(self):
        return self.response.css("title ::text").get()

    @field
    def unknown_attribute(self):
        return "foo"


EXAMPLE_RESPONSE = HttpResponse(
    "http://example.com",
    body=b"<html><head><title>Hello!</title></html>",
)


@pytest.mark.asyncio
async def test_fields() -> None:
    page = Page(response=EXAMPLE_RESPONSE)

    assert page.name == "Hello!"
    assert await page.price == "$123"

    item = await page.to_item()
    assert isinstance(item, Item)
    assert item.name == "Hello!"
    assert item.price == "$123"


@pytest.mark.asyncio
async def test_fields_invalid_page() -> None:
    page = InvalidPage(response=EXAMPLE_RESPONSE)
    with pytest.raises(
        TypeError, match="unexpected keyword argument 'unknown_attribute'"
    ):
        await page.to_item()


def test_item_from_fields_sync() -> None:
    @attrs.define
    class Page(ItemPage):
        @field
        def name(self):
            return "name"

        def to_item(self):
            return item_from_fields_sync(self, dict)

    page = Page()
    assert page.to_item() == dict(name="name")


def test_field_non_callable() -> None:
    with pytest.raises(TypeError):

        @attrs.define
        class Page(ItemPage):
            # https://github.com/python/mypy/issues/1362#issuecomment-438246775
            @field  # type: ignore
            @property
            def name(self):
                return "name"

            def to_item(self):
                return item_from_fields_sync(self, dict)


def test_field_classmethod() -> None:
    with pytest.raises(TypeError):

        @attrs.define
        class Page(ItemPage):
            @field
            @classmethod
            def name(cls):
                return "name"

            def to_item(self):
                return item_from_fields_sync(self, dict)


@pytest.mark.asyncio
async def test_field_order() -> None:
    class DictItemPage(Page):
        async def to_item(self):
            return await item_from_fields(self)

    page = DictItemPage(response=EXAMPLE_RESPONSE)
    item = await page.to_item()
    assert item == {"name": "Hello!", "price": "$123"}
    assert list(item.keys()) == ["name", "price"]


def test_field_decorator_no_arguments() -> None:
    class Page(ItemPage):
        @field()
        def name(self):
            return "Name"

        def to_item(self):
            return item_from_fields_sync(self)

    page = Page()
    assert page.to_item() == {"name": "Name"}


def test_field_cache_sync() -> None:
    class Page(ItemPage):
        _n_called_1 = 0
        _n_called_2 = 0

        def __init__(self, name):
            self.name = name

        @field(cached=True)
        def n_called_1(self):
            self._n_called_1 += 1
            return self._n_called_1, self.name

        @field(cached=False)
        def n_called_2(self):
            self._n_called_2 += 1
            return self._n_called_2, self.name

    pages = [Page("first"), Page("second")]
    for page in pages:
        assert page.n_called_1 == (1, page.name)
        assert page.n_called_1 == (1, page.name)

        assert page.n_called_2 == (1, page.name)
        assert page.n_called_2 == (2, page.name)


@pytest.mark.asyncio
async def test_field_cache_async() -> None:
    class Page(ItemPage):
        _n_called_1 = 0
        _n_called_2 = 0

        def __init__(self, name):
            self.name = name

        @field(cached=True)
        async def n_called_1(self):
            self._n_called_1 += 1
            return self._n_called_1, self.name

        @field(cached=False)
        async def n_called_2(self):
            self._n_called_2 += 1
            return self._n_called_2, self.name

    pages = [Page("first"), Page("second")]
    for page in pages:
        assert await page.n_called_1 == (1, page.name)
        assert await page.n_called_1 == (1, page.name)

        assert await page.n_called_2 == (1, page.name)
        assert await page.n_called_2 == (2, page.name)


@pytest.mark.asyncio
async def test_field_cache_async_locked() -> None:
    class Page(ItemPage):
        _n_called = 0

        @field(cached=True)
        async def n_called(self):
            await asyncio.sleep(random.randint(0, 10) / 100.0)
            self._n_called += 1
            return self._n_called

    page = Page()
    results = await asyncio.gather(
        page.n_called,
        page.n_called,
        page.n_called,
        page.n_called,
        page.n_called,
    )
    assert results == [1, 1, 1, 1, 1]


@pytest.mark.asyncio
async def test_skip_nonitem_fields_async() -> None:
    class ExtendedPage(Page):
        @field
        def new_attribute(self):
            return "foo"

    page = ExtendedPage(response=EXAMPLE_RESPONSE)
    with pytest.raises(TypeError, match="unexpected keyword argument 'new_attribute'"):
        await page.to_item()

    class ExtendedPage2(ExtendedPage):
        async def to_item(self) -> Item:
            return await item_from_fields(self, Item, skip_nonitem_fields=True)

    page = ExtendedPage2(response=EXAMPLE_RESPONSE)
    item = await page.to_item()
    assert item == Item(name="Hello!", price="$123")


def test_skip_nonitem_fields() -> None:
    @attrs.define
    class SyncPage(ItemPage):
        response: HttpResponse

        @field
        def name(self):
            return self.response.css("title ::text").get()

        @field
        def price(self):
            return "$123"

        def to_item(self) -> Item:  # type: ignore[override]
            return item_from_fields_sync(self, Item)

    class ExtendedPage(SyncPage):
        @field
        def new_attribute(self):
            return "foo"

    page = ExtendedPage(response=EXAMPLE_RESPONSE)
    with pytest.raises(TypeError, match="unexpected keyword argument 'new_attribute'"):
        page.to_item()

    class ExtendedPage2(ExtendedPage):
        def to_item(self) -> Item:  # type: ignore[override]
            return item_from_fields_sync(self, Item, skip_nonitem_fields=True)

    page = ExtendedPage2(response=EXAMPLE_RESPONSE)
    item = page.to_item()
    assert item == Item(name="Hello!", price="$123")


def test_field_meta() -> None:
    class MyPage(ItemPage):
        @field(meta={"good": True})
        def field1(self):
            return "foo"

        @field
        def field2(self):
            return "foo"

        def to_item(self):
            return item_from_fields_sync(self)

    page = MyPage()
    for fields in [get_fields_dict(MyPage), get_fields_dict(page)]:
        assert list(fields.keys()) == ["field1", "field2"]
        assert fields["field1"].name == "field1"
        assert fields["field1"].meta == {"good": True}

        assert fields["field2"].name == "field2"
        assert fields["field2"].meta is None


def test_field_subclassing() -> None:
    class Page(ItemPage):
        @field
        def field1(self):
            return 1

        @field
        def field3(self):
            return 1

    assert list(get_fields_dict(Page)) == ["field1", "field3"]
    assert get_fields_dict(Page)["field3"].meta is None

    class Page2(Page):
        @field
        def field2(self):
            return 1

        @field(meta={"foo": "bar"})
        def field3(self):
            return 1

    assert get_fields_dict(Page2)["field3"].meta == {"foo": "bar"}
    assert list(get_fields_dict(Page2)) == ["field1", "field3", "field2"]

    assert get_fields_dict(Page)["field3"].meta is None
    assert list(get_fields_dict(Page)) == ["field1", "field3"]

    class Page3(Page2):
        @field
        def field3(self):
            return 2

    assert get_fields_dict(Page3)["field3"].meta is None
    assert list(get_fields_dict(Page3)) == ["field1", "field3", "field2"]

    assert get_fields_dict(Page)["field3"].meta is None
    assert list(get_fields_dict(Page)) == ["field1", "field3"]

    assert get_fields_dict(Page2)["field3"].meta == {"foo": "bar"}
    assert list(get_fields_dict(Page2)) == ["field1", "field3", "field2"]


def test_field_subclassing_super() -> None:
    class Page(ItemPage):
        @field
        def field1(self):
            return 1

    class Page2(Page):
        @field
        def field1(self):
            return super().field1 + 1

    page = Page()
    assert page.field1 == 1
    page2 = Page2()
    assert page2.field1 == 2


def test_field_subclassing_from_to_item() -> None:
    # to_item() should be the same since it was not overridden from the
    # subclass.
    class PageToItem(ItemPage):
        def to_item(self):
            return {"field1": 1, "field2": 2, "field3": 3, "field4": 4}

    class Page1(PageToItem):
        @field
        def field1(self):
            return 0

    page_1 = Page1()
    assert page_1.field1 == 0
    assert page_1.to_item() == {"field1": 1, "field2": 2, "field3": 3, "field4": 4}

    # to_item() only reflects the field that was decorated.
    class Page2(PageToItem):
        @field
        def field2(self):
            return 0

        def to_item(self):
            return item_from_fields_sync(self)

    page_2 = Page2()
    assert page_2.field2 == 0
    assert page_2.to_item() == {"field2": 0}

    # to_item() raises an error if there are some required fields from the item_cls
    # that doesn't have a corresponding field value.
    @attrs.define
    class SomeItem:
        field1: int
        field2: int
        field3: int
        field4: int

    class Page3(PageToItem):
        @field
        def field3(self):
            return 0

        def to_item(self):
            return item_from_fields_sync(self, item_cls=SomeItem)

    page_3 = Page3()
    assert page_3.field3 == 0
    with pytest.raises(TypeError):
        page_3.to_item()


def test_field_with_other_decorators() -> None:
    def clean_str(method):
        def wrapper(*args, **kwargs):
            return method(*args, **kwargs).strip()

        return wrapper

    class MyPage(ItemPage):
        @field
        @clean_str
        def field_foo(self):
            return " foo  \n"

        @field(meta={"good": True})
        @clean_str
        def field_foo_meta(self):
            return " foo  \n"

        @field(cached=True)
        @clean_str
        def field_foo_cached(self):
            return " foo  \n"

    page = MyPage()
    assert page.field_foo == "foo"
    assert page.field_foo_meta == "foo"
    assert page.field_foo_cached == "foo"


@pytest.mark.asyncio
async def test_field_with_handle_urls() -> None:
    page = ProductPage()
    assert page.name == "name"
    assert page.price == 12.99
    assert await page.to_item() == Product(name="name", price=12.99)

    page = ImprovedProductPage()
    assert page.name == "improved name"
    assert page.price == 12.99
    assert await page.to_item() == Product(name="improved name", price=12.99)

    page = SimilarProductPage()
    assert page.name == "name"
    assert page.price == 12.99
    assert await page.to_item() == ProductSimilar(name="name", price=12.99)

    page = MoreProductPage()
    assert page.name == "name"
    assert page.price == 12.99
    assert page.brand == "brand"
    assert await page.to_item() == ProductMoreFields(
        name="name", price=12.99, brand="brand"
    )

    page = LessProductPage()
    assert page.name == "name"
    assert await page.to_item() == ProductFewerFields(name="name")

    for page in [  # type: ignore[assignment]
        CustomProductPage(),
        CustomProductPageNoReturns(),
        CustomProductPageDataTypeOnly(),
    ]:
        assert page.name == "name"
        assert page.price == 12.99
        assert await page.to_item() == Product(name="name", price=12.99)


def test_field_processors_sync() -> None:
    def proc1(s):
        return s + "x"

    @attrs.define
    class Page(ItemPage):
        @field(out=[str.strip, proc1])
        def name(self):
            return "  name\t "

    page = Page()
    assert page.name == "namex"


@pytest.mark.asyncio
async def test_field_processors_async() -> None:
    def proc1(s):
        return s + "x"

    @attrs.define
    class Page(ItemPage):
        @field(out=[str.strip, proc1])
        async def name(self):
            return "  name\t "

    page = Page()
    assert await page.name == "namex"


def test_field_processors_inheritance() -> None:
    def proc1(s):
        return s + "x"

    class BasePage(ItemPage):
        @field(out=[str.strip, proc1])
        def name(self):
            return "  name\t "

    class Page(BasePage):
        @field(out=[str.strip])
        def name(self):
            return "  name\t "

    base_page = BasePage()
    assert base_page.name == "namex"
    page = Page()
    assert page.name == "name"


def test_field_processors_page() -> None:
    def proc1(s, page):
        return page.prefix + s + "x"

    class Page(ItemPage):
        @field(out=[str.strip, proc1])
        def name(self):
            return "  name\t "

        @field
        def prefix(self):
            return "prefix: "

    page = Page()
    assert page.name == "prefix: namex"


def test_field_processors_multiple_pages() -> None:
    def proc(value, page):
        return page.body + value

    class Page(WebPage):
        @field
        def body(self):
            return self.response.text

        @field(out=[proc])
        def processed(self):
            return "suffix"

    page1 = Page(response=HttpResponse("https://example.com", b"page1"))
    page2 = Page(response=HttpResponse("https://example.com", b"page2"))
    assert page1.body == "page1"
    assert page1.processed == "page1suffix"
    assert page2.body == "page2"
    assert page2.processed == "page2suffix"


def test_field_processors_circular() -> None:
    def proc1(s, page):
        return s + page.b

    def proc2(s, page):
        return s + page.a

    class Page(ItemPage):
        @field(out=[proc1])
        def a(self):
            return "a"

        @field(out=[proc2])
        def b(self):
            return "b"

    page = Page()
    with pytest.raises(RecursionError):
        page.a
    with pytest.raises(RecursionError):
        page.b


def test_field_processors_default() -> None:
    @attrs.define
    class BasePage(ItemPage):
        class Processors:
            name = [str.strip]

        @field
        def name(self):
            return "  name\t "

    class Page(BasePage):
        pass

    base_page = BasePage()
    assert base_page.name == "name"

    page = Page()
    assert page.name == "name"


def test_field_processors_override() -> None:
    def proc1(s):
        return s + "x"

    class BasePage(ItemPage):
        class Processors:
            f1: List[Callable] = [str.strip]
            f2 = [str.strip]
            f3 = [str.strip]
            f4: List[Callable] = [str.strip]
            f5: List[Callable] = [str.strip]

        @field
        def f1(self):
            return "  f1\t "

        @field(out=[])
        def f2(self):
            return "  f2\t "

        @field
        def f3(self):
            return "  f3\t "

        @field
        def f4(self):
            return "  f4\t "

        @field
        def f5(self):
            return "  f5\t "

    class Page(BasePage):
        class Processors(BasePage.Processors):
            f1 = [proc1]
            f4 = BasePage.Processors.f4 + [proc1]

        @field(out=BasePage.Processors.f5 + [proc1])
        def f5(self):
            return "  f5\t "

    base_page = BasePage()
    assert base_page.f1 == "f1"
    assert base_page.f2 == "  f2\t "
    assert base_page.f3 == "f3"
    assert base_page.f4 == "f4"
    assert base_page.f5 == "f5"

    page = Page()
    assert page.f1 == "  f1\t x"
    assert page.f2 == "  f2\t "
    assert page.f3 == "f3"
    assert page.f4 == "f4x"
    assert page.f5 == "f5x"


def test_field_processors_super() -> None:
    class BasePage(ItemPage):
        class Processors:
            name = [str.strip]
            desc = [str.strip]

        @field
        def name(self):
            return "name "

        @field
        def desc(self):
            return "desc "

    class Page(BasePage):
        class Processors(BasePage.Processors):
            name: List[Callable] = []

        @field
        def name(self):
            base_name = super().name
            return base_name + "2 "

    class Page2(Page):
        class Processors(Page.Processors):
            name: List[Callable] = []
            desc: List[Callable] = []

        @field
        def desc(self):
            base_desc = super().desc
            return base_desc + "2 "

    base_page = BasePage()
    assert base_page.name == "name"
    page = Page()
    assert page.name == "name 2 "
    page2 = Page2()
    assert page2.desc == "desc 2 "


def test_field_processors_builtin() -> None:
    @attrs.define
    class Page(ItemPage):
        @field(out=[int])
        def value(self):
            return "1"

    page = Page()
    assert page.value == 1


def test_field_mixin() -> None:
    class A(ItemPage):
        @field
        def a(self):
            return None

    class Mixin(FieldsMixin):
        @field
        def mixin(self):
            return None

    class B(Mixin, A):
        @field
        def b(self):
            return None

    class C(Mixin, A):
        @field
        def c(self):
            return None

    assert set(get_fields_dict(A)) == {"a"}
    assert set(get_fields_dict(B)) == {"a", "b", "mixin"}
    assert set(get_fields_dict(C)) == {"a", "c", "mixin"}



from url_matcher import Patterns

from tests.po_lib import POBase
from web_poet import ItemPage, handle_urls


class PONestedPkgOverriden(ItemPage): ...


@handle_urls(
    include=["example.com", "example.org"],
    exclude=["/*.jpg|"],
    instead_of=PONestedPkgOverriden,
)
class PONestedPkg(POBase):
    expected_instead_of = PONestedPkgOverriden
    expected_patterns = Patterns(["example.com", "example.org"], ["/*.jpg|"])
    expected_to_return = None
    expected_meta = {}  # type: ignore



import sys
from typing import Type

import attrs
import pytest

from web_poet import (
    HttpClient,
    HttpResponse,
    HttpResponseBody,
    Injectable,
    ItemPage,
    PageParams,
    ResponseUrl,
    Stats,
    WebPage,
)
from web_poet.annotated import AnnotatedInstance
from web_poet.page_inputs.url import _Url
from web_poet.serialization import (
    SerializedDataFileStorage,
    SerializedLeafData,
    deserialize,
    deserialize_leaf,
    register_serialization,
    serialize,
    serialize_leaf,
)


def _assert_pages_equal(p1, p2) -> None:
    assert type(p1) == type(p2)  # noqa: E721
    assert type(p1.response) == type(p2.response)  # noqa: E721
    assert type(p1.response.body) == type(p2.response.body)  # noqa: E721
    assert type(p1.response.headers) == type(p2.response.headers)  # noqa: E721
    assert p1.response.body == p2.response.body
    assert p1.response.status == p2.response.status
    assert p1.response.headers == p2.response.headers
    assert p1.response._encoding == p2.response._encoding
    _assert_urls_equal(p1.response.url, p2.response.url)


def _assert_urls_equal(u1: _Url, u2: _Url) -> None:
    assert type(u1) == type(u2)  # noqa: E721
    assert str(u1) == str(u2)


def test_serialization_leaf() -> None:
    leaf = HttpResponseBody(b"foo")
    serialized_data = serialize_leaf(leaf)
    assert isinstance(serialized_data["html"], bytes)
    assert HttpResponseBody(serialized_data["html"]) == leaf
    deserialized_data = deserialize_leaf(HttpResponseBody, serialized_data)
    assert leaf == deserialized_data


def test_serialization_leaf_unsupported() -> None:
    class A:
        pass

    with pytest.raises(
        NotImplementedError, match=r"Serialization .+ is not implemented"
    ):
        serialize_leaf(A())

    with pytest.raises(
        NotImplementedError, match=r"Deserialization .+ is not implemented"
    ):
        deserialize_leaf(A, {})


def test_serialization(book_list_html_response) -> None:
    @attrs.define
    class ResponseData(Injectable):
        response: HttpResponse

    @attrs.define
    class MyWebPage(ItemPage):
        response: HttpResponse
        url: ResponseUrl
        params: PageParams
        data: ResponseData
        stats: Stats

    url_str = "http://books.toscrape.com/index.html"
    url = ResponseUrl(url_str)
    page_params = PageParams(foo="bar")
    stats = Stats()

    serialized_deps = serialize([book_list_html_response, url, page_params, stats])
    info_json = f"""{{
  "_encoding": "utf-8",
  "headers": [],
  "status": null,
  "url": "{url_str}"
}}""".encode()
    assert serialized_deps == {
        "HttpResponse": {
            "body.html": bytes(book_list_html_response.body),
            "info.json": info_json,
        },
        "ResponseUrl": {
            "txt": url_str.encode(),
        },
        "PageParams": {
            "json": b'{\n  "foo": "bar"\n}',
        },
        "Stats": {},
    }

    po = MyWebPage(
        book_list_html_response,
        url,
        page_params,
        ResponseData(book_list_html_response),
        Stats(),
    )
    deserialized_po = deserialize(MyWebPage, serialized_deps)
    _assert_pages_equal(po, deserialized_po)
    assert deserialized_po.data is not None


def test_serialization_injectable(book_list_html_response) -> None:
    with pytest.raises(ValueError, match=r"Injectable type .+ passed"):
        serialize([WebPage(book_list_html_response)])


def test_serialization_httpresponse_encoding(book_list_html) -> None:
    body = HttpResponseBody(bytes(book_list_html, "utf-8"))
    resp_enc = HttpResponse(
        url=ResponseUrl("http://books.toscrape.com/index.html"),
        body=body,
        encoding="utf-8",
    )
    assert resp_enc._encoding == "utf-8"
    deserialized_resp_enc = deserialize_leaf(HttpResponse, serialize_leaf(resp_enc))
    assert deserialized_resp_enc._encoding == "utf-8"

    resp_noenc = HttpResponse(
        url=ResponseUrl("http://books.toscrape.com/index.html"), body=body
    )
    assert resp_noenc._encoding is None
    deserialized_resp_noenc = deserialize_leaf(HttpResponse, serialize_leaf(resp_noenc))
    assert deserialized_resp_noenc._encoding is None


def test_custom_functions() -> None:
    class C:
        value: int

        def __init__(self, value: int):
            self.value = value

    def _serialize(o: C) -> SerializedLeafData:
        return {"bin": o.value.to_bytes((o.value.bit_length() + 7) // 8, "little")}

    def _deserialize(t: Type[C], data: SerializedLeafData) -> C:
        return t(int.from_bytes(data["bin"], "little"))

    register_serialization(_serialize, _deserialize)

    obj = C(22222222222)
    deserialized_obj = deserialize_leaf(C, serialize_leaf(obj))
    assert obj.value == deserialized_obj.value


def test_write_data(book_list_html_response, tmp_path) -> None:
    @attrs.define
    class MyWebPage(ItemPage):
        response: HttpResponse
        url: ResponseUrl

    url = ResponseUrl("http://example.com")

    directory = tmp_path / "ser"
    directory.mkdir()
    storage = SerializedDataFileStorage(directory)
    serialized_deps = serialize([book_list_html_response, url])
    storage.write(serialized_deps)
    assert (directory / "HttpResponse-body.html").exists()
    assert (directory / "HttpResponse-body.html").read_bytes() == bytes(
        book_list_html_response.body
    )
    assert (directory / "HttpResponse-info.json").exists()
    assert (directory / "ResponseUrl.txt").exists()
    assert (directory / "ResponseUrl.txt").read_text(
        encoding="utf-8"
    ) == "http://example.com"

    read_serialized_deps = storage.read()
    po = MyWebPage(book_list_html_response, url)
    deserialized_po = deserialize(MyWebPage, read_serialized_deps)
    assert type(deserialized_po) == MyWebPage  # noqa: E721
    _assert_pages_equal(po, deserialized_po)


def test_extra_files(book_list_html_response, tmp_path) -> None:
    directory = tmp_path / "ser"
    directory.mkdir()
    storage = SerializedDataFileStorage(directory)
    serialized_deps = serialize([book_list_html_response])
    storage.write(serialized_deps)
    (directory / "foo.dir").mkdir()
    (directory / "bar.txt").touch()
    read_serialized_deps = storage.read()
    assert "HttpResponse" in read_serialized_deps


def test_httpclient_empty(tmp_path) -> None:
    directory = tmp_path / "ser"
    directory.mkdir()
    client = HttpClient()
    storage = SerializedDataFileStorage(directory)
    serialized_deps = serialize([client])
    storage.write(serialized_deps)
    assert (directory / "HttpClient.exists").exists()
    read_serialized_deps = storage.read()
    assert "HttpClient" in read_serialized_deps


@pytest.mark.skipif(
    sys.version_info < (3, 9), reason="No Annotated support in Python < 3.9"
)
def test_annotated(book_list_html_response) -> None:
    from typing import Annotated

    @attrs.define
    class MyWebPage(ItemPage):
        response: Annotated[HttpResponse, "foo", 42]
        url: ResponseUrl

    url_str = "http://books.toscrape.com/index.html"
    url = ResponseUrl(url_str)

    serialized_deps = serialize(
        [AnnotatedInstance(book_list_html_response, ("foo", 42)), url]
    )
    po = MyWebPage(
        book_list_html_response,
        url,
    )
    deserialized_po = deserialize(MyWebPage, serialized_deps)
    _assert_pages_equal(po, deserialized_po)


@pytest.mark.skipif(
    sys.version_info < (3, 9), reason="No Annotated support in Python < 3.9"
)
def test_annotated_duplicate(book_list_html_response) -> None:
    url_str = "http://books.toscrape.com/index.html"
    url = ResponseUrl(url_str)

    with pytest.raises(
        ValueError, match="Several instances of AnnotatedInstance for HttpResponse were"
    ):
        serialize(
            [
                AnnotatedInstance(book_list_html_response, ("foo", 42)),
                AnnotatedInstance(book_list_html_response, ("bar",)),
                url,
            ]
        )



class PageParams(dict):
    """Container class that could contain any arbitrary data to be passed into
    a Page Object.

    Note that this is simply a subclass of Python's ``dict``.
    """

    pass



"""
HTTP Exceptions
~~~~~~~~~~~~~~~

These are exceptions pertaining to common issues faced when executing HTTP
operations.
"""

from typing import Optional

from web_poet.page_inputs.http import HttpRequest, HttpResponse


class HttpError(IOError):
    """Indicates that an exception has occurred when handling an HTTP operation.

    This is used as a **base class** for more specific errors and could be vague
    since it could denote problems either in the HTTP Request or Response.

    For more specific errors, it would be better to use :class:`.HttpRequestError`
    and :class:`.HttpResponseError`.

    :param request: Request that triggered the exception.
    :type request: HttpRequest
    """

    def __init__(self, msg: str = None, request: HttpRequest = None):
        #: Request that triggered the exception.
        self.request: Optional[HttpRequest] = request
        if msg is None:
            msg = f"An Error ocurred when executing this HTTP Request: {self.request}"
        super().__init__(msg)


class HttpRequestError(HttpError):
    """Indicates that an exception has occurred when the **HTTP Request** was
    being handled.

    :param request: The :class:`~.HttpRequest` instance that was used.
    :type request: HttpRequest
    """

    pass


class HttpResponseError(HttpError):
    """Indicates that an exception has occurred when the **HTTP Response** was
    received.

    For responses that are in the status code ``100-3xx range``, this exception
    shouldn't be raised at all. However, for responses in the ``400-5xx``, this
    will be raised by **web-poet**.

    .. note::

        Frameworks implementing **web-poet** should **NOT** raise this  exception.

        This exception is raised by web-poet itself, based on ``allow_status``
        parameter found in the methods of :class:`~.HttpClient`.

    :param request: Request that got the response that triggered the exception.
    :type request: HttpRequest
    :param response: Response that triggered the exception.
    :type response: HttpResponse
    """

    def __init__(
        self,
        msg: str = None,
        response: HttpResponse = None,
        request: HttpRequest = None,
    ):
        #: Response that triggered the exception.
        self.response: Optional[HttpResponse] = response
        if msg is None:
            msg = f"Unexpected HTTP Response received: {self.response}"
        super().__init__(msg, request=request)



"""
``web_poet.fields`` is a module with helpers for putting extraction logic
into separate Page Object methods / properties.
"""

import inspect
from contextlib import suppress
from functools import update_wrapper, wraps
from typing import Callable, Dict, List, Optional, Tuple, Type, TypeVar

import attrs
from itemadapter import ItemAdapter

from web_poet.utils import cached_method, callable_has_parameter, ensure_awaitable

_FIELDS_INFO_ATTRIBUTE_READ = "_web_poet_fields_info"
_FIELDS_INFO_ATTRIBUTE_WRITE = "_web_poet_fields_info_temp"
_FIELD_METHODS_ATTRIBUTE = "_web_poet_field_methods"


@attrs.define
class FieldInfo:
    """Information about a field"""

    #: name of the field
    name: str

    #: field metadata
    meta: Optional[dict] = None

    #: field processors
    out: Optional[List[Callable]] = None


class FieldsMixin:
    """A mixin which is required for a class to support fields"""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # To support fields, we must ensure that fields dict is not shared
        # between subclasses, i.e. a decorator in a subclass doesn't affect
        # the base class. This is done by making decorator write to a
        # temporary location, and then merging it all on subclass creation.
        this_class_fields = getattr(cls, _FIELDS_INFO_ATTRIBUTE_WRITE, {})
        base_fields = {}
        for base_class in cls.__bases__:
            fields = getattr(base_class, _FIELDS_INFO_ATTRIBUTE_READ, {})
            base_fields.update(fields)
        if base_fields or this_class_fields:
            fields = {**base_fields, **this_class_fields}
            setattr(cls, _FIELDS_INFO_ATTRIBUTE_READ, fields)
            with suppress(AttributeError):
                delattr(cls, _FIELDS_INFO_ATTRIBUTE_WRITE)
        setattr(cls, _FIELD_METHODS_ATTRIBUTE, {})


def field(
    method=None,
    *,
    cached: bool = False,
    meta: Optional[dict] = None,
    out: Optional[List[Callable]] = None,
):
    """
    Page Object method decorated with ``@field`` decorator becomes a property,
    which is then used by :class:`~.ItemPage`'s to_item() method to populate
    a corresponding item attribute.

    By default, the value is computed on each property access.
    Use ``@field(cached=True)`` to cache the property value.

    The ``meta`` parameter allows to store arbitrary information for the field,
    e.g. ``@field(meta={"expensive": True})``. This information can be later
    retrieved for all fields using the :func:`get_fields_dict` function.

    The ``out`` parameter is an optional list of field processors, which are
    functions applied to the value of the field before returning it.
    """

    class _field:
        def __init__(self, method):
            if not callable(method):
                raise TypeError(
                    f"@field decorator must be used on methods, {method!r} is decorated instead"
                )
            self.original_method = method
            self.name: Optional[str] = None

        def __set_name__(self, owner, name):
            self.name = name
            if not hasattr(owner, _FIELDS_INFO_ATTRIBUTE_WRITE):
                setattr(owner, _FIELDS_INFO_ATTRIBUTE_WRITE, {})

            field_info = FieldInfo(name=name, meta=meta, out=out)
            getattr(owner, _FIELDS_INFO_ATTRIBUTE_WRITE)[name] = field_info

        def __get__(self, instance, owner=None):
            # We use the original method and the out arg from the field and
            # the Processors class from the instance class, so caching needs to
            # take into account the instance class and the field object. So we
            # use the field object id() as a key when caching the method in
            # the instance class.
            cache_key = id(self)
            method = self._get_processed_method(owner, cache_key)
            if method is None:
                if out is not None:
                    processor_functions = out
                elif hasattr(owner, "Processors"):
                    processor_functions = getattr(owner.Processors, self.name, [])
                else:
                    processor_functions = []
                processors: List[Tuple[Callable, bool]] = []
                for processor_function in processor_functions:
                    takes_page = callable_has_parameter(processor_function, "page")
                    processors.append((processor_function, takes_page))
                method = self._processed(self.original_method, processors)
                if cached:
                    method = cached_method(method)
                self._set_processed_method(owner, cache_key, method)

            return method(instance)

        @staticmethod
        def _get_processed_method(cls, key):
            return getattr(cls, _FIELD_METHODS_ATTRIBUTE).get(key)

        @staticmethod
        def _set_processed_method(cls, key, method):
            getattr(cls, _FIELD_METHODS_ATTRIBUTE)[key] = method

        @staticmethod
        def _process(value, page, processors):
            for processor, takes_page in processors:
                if takes_page:
                    value = processor(value, page=page)
                else:
                    value = processor(value)
            return value

        @staticmethod
        def _processed(method, processors):
            """Returns a wrapper for method that calls processors on its result"""
            if inspect.iscoroutinefunction(method):

                async def processed(page):
                    if hasattr(page, "_validate_input"):
                        validation_item = page._validate_input()
                        if validation_item is not None:
                            return getattr(validation_item, method.__name__)
                    return _field._process(await method(page), page, processors)

            else:

                def processed(page):
                    if hasattr(page, "_validate_input"):
                        validation_item = page._validate_input()
                        if validation_item is not None:
                            return getattr(validation_item, method.__name__)
                    return _field._process(method(page), page, processors)

            return wraps(method)(processed)

    if method is not None:
        # @field syntax
        res = _field(method)
        update_wrapper(res, method)
        return res
    else:
        # @field(...) syntax
        return _field


def get_fields_dict(cls_or_instance) -> Dict[str, FieldInfo]:
    """Return a dictionary with information about the fields defined
    for the class: keys are field names, and values are
    :class:`web_poet.fields.FieldInfo` instances.
    """
    return getattr(cls_or_instance, _FIELDS_INFO_ATTRIBUTE_READ, {})


T = TypeVar("T")


# FIXME: type is ignored as a workaround for https://github.com/python/mypy/issues/3737
# inference works properly if a non-default item_cls is passed; for dict
# it's not working (return type is Any)
async def item_from_fields(
    obj, item_cls: Type[T] = dict, *, skip_nonitem_fields: bool = False  # type: ignore[assignment]
) -> T:
    """Return an item of ``item_cls`` type, with its attributes populated
    from the ``obj`` methods decorated with :class:`field` decorator.

    If ``skip_nonitem_fields`` is True, ``@fields`` whose names are not
    among ``item_cls`` field names are not passed to ``item_cls.__init__``.

    When ``skip_nonitem_fields`` is False (default), all ``@fields`` are passed
    to ``item_cls.__init__``, possibly causing exceptions if
    ``item_cls.__init__`` doesn't support them.
    """
    item_dict = item_from_fields_sync(obj, item_cls=dict, skip_nonitem_fields=False)
    field_names = list(item_dict.keys())
    if skip_nonitem_fields:
        field_names = _without_unsupported_field_names(item_cls, field_names)
    return item_cls(
        **{name: await ensure_awaitable(item_dict[name]) for name in field_names}
    )


def item_from_fields_sync(
    obj, item_cls: Type[T] = dict, *, skip_nonitem_fields: bool = False  # type: ignore[assignment]
) -> T:
    """Synchronous version of :func:`item_from_fields`."""
    field_names = list(get_fields_dict(obj))
    if skip_nonitem_fields:
        field_names = _without_unsupported_field_names(item_cls, field_names)
    return item_cls(**{name: getattr(obj, name) for name in field_names})


def _without_unsupported_field_names(
    item_cls: type, field_names: List[str]
) -> List[str]:
    item_field_names = ItemAdapter.get_field_names_from_class(item_cls)
    if item_field_names is None:  # item_cls doesn't define field names upfront
        return field_names[:]
    return list(set(field_names) & set(item_field_names))



from url_matcher import Patterns

from tests.po_lib import POBase
from web_poet import ItemPage, handle_urls


class POModuleOverriden(ItemPage): ...


@handle_urls("example.com", instead_of=POModuleOverriden, extra_arg="foo")
class POModule(POBase):
    expected_instead_of = POModuleOverriden
    expected_patterns = Patterns(["example.com"])
    expected_to_return = None
    expected_meta = {"extra_arg": "foo"}  # type: ignore



from abc import ABC, abstractmethod
from typing import Any, Dict, Union

StatNum = Union[int, float]


class StatCollector(ABC):
    """Base class for web-poet to implement the storing of data written through
    :class:`~web_poet.page_inputs.stats.Stats`."""

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        """Set the value of stat *key* to *value*."""

    @abstractmethod
    def inc(self, key: str, value: StatNum = 1) -> None:
        """Increment the value of stat *key* by *value*, or set it to *value*
        if *key* has no value."""


class DummyStatCollector(StatCollector):
    """:class:`~web_poet.page_inputs.stats.StatCollector` implementation that
    does not persist stats. It is used when running automatic tests, where stat
    storage is not necessary."""

    def __init__(self):
        self._stats: Dict[str, Any] = {}

    def set(self, key: str, value: Any) -> None:  # noqa: D102
        self._stats[key] = value

    def inc(self, key: str, value: StatNum = 1) -> None:  # noqa: D102
        if key in self._stats:
            assert isinstance(self._stats[key], (int, float))
            self._stats[key] += value
        else:
            self._stats[key] = value


class Stats:
    """Page input class to write key-value data pairs during parsing that you
    can inspect later. See :ref:`stats`.

    Stats can be set to a fixed value or, if numeric, incremented.

    Stats are write-only.

    Storage and read access of stats depends on the web-poet framework that you
    are using. Check the documentation of your web-poet framework to find out
    if it supports stats, and if so, how to read stored stats.
    """

    def __init__(self, stat_collector=None):
        self._stats = stat_collector or DummyStatCollector()

    def set(self, key: str, value: Any) -> None:
        """Set the value of stat *key* to *value*."""
        self._stats.set(key, value)

    def inc(self, key: str, value: StatNum = 1) -> None:
        """Increment the value of stat *key* by *value*, or set it to *value*
        if *key* has no value."""
        self._stats.inc(key, value)



from __future__ import annotations  # https://www.python.org/dev/peps/pep-0563/

import importlib
import importlib.util
import pkgutil
import warnings
from collections import defaultdict, deque
from operator import attrgetter
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

import attrs
from url_matcher import Patterns, URLMatcher

from web_poet.page_inputs.url import _Url
from web_poet.pages import ItemPage, get_item_cls
from web_poet.utils import _create_deprecated_class, as_list, str_to_pattern

Strings = Union[str, Iterable[str]]

RulesRegistryTV = TypeVar("RulesRegistryTV", bound="RulesRegistry")


@attrs.define(frozen=True)
class ApplyRule:
    """A rule that primarily applies Page Object and Item overrides for a given
    URL pattern.

    This is instantiated when using the :func:`web_poet.handle_urls` decorator.
    It's also being returned as a ``List[ApplyRule]`` when calling the
    ``web_poet.default_registry``'s :meth:`~.RulesRegistry.get_rules`
    method.

    You can access any of its attributes:

        * ``for_patterns`` - contains the list of URL patterns associated with
          this rule. You can read the API documentation of the `url-matcher
          <https://url-matcher.readthedocs.io/>`_ package for more information
          about the patterns.
        * ``use`` - The Page Object that will be **used** in cases where the URL
          pattern represented by the ``for_patterns`` attribute is matched.
        * ``instead_of`` - *(optional)* The Page Object that will be **replaced**
          with the Page Object specified via the ``use`` parameter.
        * ``to_return`` - *(optional)* The item class that the Page Object specified
          in ``use`` is capable of returning.
        * ``meta`` - *(optional)* Any other information you may want to store.
          This doesn't do anything for now but may be useful for future API updates.

    The main functionality of this class lies in the ``instead_of`` and ``to_return``
    parameters. Should both of these be omitted, then :class:`~.ApplyRule` simply
    tags which URL patterns the given Page Object defined in ``use`` is expected
    to be used on.

    When ``to_return`` is not None (e.g. ``to_return=MyItem``),
    the Page Object in ``use`` is declared as capable of returning a certain
    item class (i.e. ``MyItem``).

    When ``instead_of`` is not None (e.g. ``instead_of=ReplacedPageObject``),
    the rule adds an expectation that the ``ReplacedPageObject`` wouldn't
    be used for the URLs matching ``for_patterns``, since the Page Object
    in ``use`` will replace it.

    If there are multiple rules which match a certain URL, the rule
    to apply is picked based on the priorities set in ``for_patterns``.

    More information regarding its usage in :ref:`rules`.

    .. tip::

        The :class:`~.ApplyRule` is also hashable. This makes it easy to store
        unique rules and identify any duplicates.
    """

    for_patterns: Patterns = attrs.field(converter=str_to_pattern)
    use: Type[ItemPage] = attrs.field(kw_only=True)
    instead_of: Optional[Type[ItemPage]] = attrs.field(default=None, kw_only=True)
    to_return: Optional[Type[Any]] = attrs.field(default=None, kw_only=True)
    meta: Dict[str, Any] = attrs.field(factory=dict, kw_only=True)

    def __hash__(self):
        return hash((self.for_patterns, self.use, self.instead_of, self.to_return))


class RulesRegistry:
    """
    RulesRegistry provides features for storing, retrieving,
    and searching for the :class:`~.ApplyRule` instances.

    ``web-poet`` provides a default Registry named ``default_registry``
    for convenience. It can be accessed this way:

    .. code-block:: python

        from web_poet import handle_urls, default_registry, WebPage
        from my_items import Product

        @handle_urls("example.com")
        class ExampleComProductPage(WebPage[Product]):
            ...

        rules = default_registry.get_rules()

    The ``@handle_urls`` decorator exposed as ``web_poet.handle_urls`` is a
    shortcut for ``default_registry.handle_urls``.

    .. note::

        It is encouraged to use the ``web_poet.default_registry`` instead of
        creating your own :class:`~.RulesRegistry` instance. Using multiple
        registries would be unwieldy in most cases.

        However, it might be applicable in certain scenarios like storing custom
        rules to separate it from the ``default_registry``.
    """

    def __init__(self, *, rules: Optional[Iterable[ApplyRule]] = None):
        self._rules: Dict[int, ApplyRule] = {}
        self._overrides_matchers: DefaultDict[Optional[Type[ItemPage]], URLMatcher] = (
            defaultdict(URLMatcher)
        )
        self._item_matchers: DefaultDict[Optional[Type], URLMatcher] = defaultdict(
            URLMatcher
        )

        # Ensures that URLMatcher is deterministic in returning a rule when
        # matching. As of url_macher==0.2.0, `url_matcher.URLMatcher._sort_domain`
        # has this sorting criteria:
        #   * Priority (descending)
        #   * Sorted list of includes for this domain (descending)
        #   * Rule identifier (descending)
        # This means that if the priority and domain are the same, the last tie
        # breaker would be the "Rule identifier", this means we can base it on
        # the order of rule addition to the registry, i.e. a counter.
        self._rule_counter = 0

        if rules is not None:
            for rule in rules:
                self.add_rule(rule)

    def add_rule(self, rule: ApplyRule) -> None:
        """Registers an :class:`web_poet.rules.ApplyRule` instance."""

        matched = self._item_matchers.get(rule.to_return)
        if matched:
            # A common case when a page object subclasses another one with the
            # same URL pattern.
            pattern_dupes = {
                pattern
                for pattern in matched.patterns.values()
                if pattern == rule.for_patterns
            }
            if pattern_dupes:
                rules_to_warn = [
                    r
                    for p in pattern_dupes
                    for r in self.search(for_patterns=p, to_return=rule.to_return)
                ] + [rule]
                warnings.warn(
                    f"The registry contains {len(rules_to_warn)} conflicting "
                    f"rules with to_return={rule.to_return} "
                    f"and the same URL pattern:\n\n"
                    f"{self._format_list(pattern_dupes)} "
                    f"\n\n"
                    f"The first rule added to the registry is used when the URL patterns are the same and "
                    f"the priorities are equal; other rules are ignored. "
                    f"This is error-prone. Consider setting the priority explicitly "
                    f"for these rules:\n\n"
                    f"{self._format_list(rules_to_warn)}",
                    stacklevel=3,  # optimized for the common case of @handle_urls
                )

        self._rule_counter += 1
        rule_id = self._rule_counter

        self._overrides_matchers[rule.instead_of].add_or_update(
            rule_id, rule.for_patterns
        )
        self._item_matchers[rule.to_return].add_or_update(rule_id, rule.for_patterns)

        self._rules[rule_id] = rule

    @classmethod
    def _format_list(cls, objects: Iterable) -> str:
        return "\n".join(repr(rule) for rule in objects)

    @classmethod
    def from_override_rules(
        cls: Type[RulesRegistryTV], rules: List[ApplyRule]
    ) -> RulesRegistryTV:
        """Deprecated. Use ``RulesRegistry(rules=...)`` instead."""
        msg = (
            "The 'from_override_rules' method is deprecated. "
            "Use 'RulesRegistry(rules=...)' instead."
        )
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return cls(rules=rules)

    def handle_urls(
        self,
        include: Strings,
        *,
        overrides: Optional[Type[ItemPage]] = None,
        instead_of: Optional[Type[ItemPage]] = None,
        to_return: Optional[Type] = None,
        exclude: Optional[Strings] = None,
        priority: int = 500,
        **kwargs,
    ):
        """
        Class decorator that indicates that the decorated Page Object should work
        for the given URL patterns.

        The URL patterns are matched using the ``include`` and ``exclude``
        parameters while ``priority`` breaks any ties. See the documentation
        of the `url-matcher <https://url-matcher.readthedocs.io/>`_ package for
        more information about them.

        This decorator is able to derive the item class returned by the Page
        Object. This is important since it marks what type of item the Page
        Object is capable of returning for the given URL patterns. For certain
        advanced cases, you can pass a ``to_return`` parameter which replaces
        any derived values (though this isn't generally recommended).

        Passing another Page Object into the ``instead_of`` parameter indicates
        that the decorated Page Object will be used instead of that for the given
        set of URL patterns. See :ref:`rule-precedence`.

        Any extra parameters are stored as meta information that can be later used.

        :param include: The URLs that should be handled by the decorated Page Object.
        :param instead_of: The Page Object that should be `replaced`.
        :param to_return: The item class holding the data returned by the Page Object.
            This could be omitted as it could be derived from the ``Returns[ItemClass]``
            or ``ItemPage[ItemClass]`` declaration of the Page Object. See
            :ref:`item-classes` section.
        :param exclude: The URLs for which the Page Object should **not** be applied.
        :param priority: The resolution priority in case of `conflicting` rules.
            A conflict happens when the ``include``, ``override``, and ``exclude``
            parameters are the same. If so, the `highest priority` will be
            chosen.
        """

        def wrapper(cls):
            if overrides is not None:
                msg = (
                    "The 'overrides' parameter in @handle_urls is deprecated. "
                    "Use the 'instead_of' parameter instead. If both 'instead_of' "
                    "and 'overrides' are provided, the latter is ignored."
                )
                warnings.warn(msg, DeprecationWarning, stacklevel=2)

            rule = ApplyRule(
                for_patterns=Patterns(
                    include=as_list(include),
                    exclude=as_list(exclude),
                    priority=priority,
                ),
                use=cls,
                instead_of=instead_of or overrides,
                to_return=to_return or get_item_cls(cls),
                meta=kwargs,
            )
            self.add_rule(rule)
            return cls

        return wrapper

    def get_rules(self) -> List[ApplyRule]:
        """Return all the :class:`~.ApplyRule` that were declared using
        the ``@handle_urls`` decorator.

        .. note::

            Remember to consider calling :func:`~.web_poet.rules.consume_modules`
            beforehand to recursively import all submodules which contains the
            ``@handle_urls`` decorators from external Page Objects.
        """
        return list(self._rules.values())

    def get_overrides(self) -> List[ApplyRule]:
        """Deprecated, use :meth:`~.RulesRegistry.get_rules` instead."""
        msg = "The 'get_overrides' method is deprecated. Use 'get_rules' instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.get_rules()

    def search(self, **kwargs) -> List[ApplyRule]:
        """Return any :class:`ApplyRule` from the registry that matches with all
        the provided attributes.

        Sample usage:

        .. code-block:: python

            rules = registry.search(use=ProductPO, instead_of=GenericPO)
            print(len(rules))           # 1
            print(rules[0].use)         # ProductPO
            print(rules[0].instead_of)  # GenericPO

        """
        # Use a dict instead of set() to preserve the order.
        rule_ids = {}

        if "to_return" in kwargs:
            matcher = self._item_matchers.get(kwargs["to_return"])
            if matcher:
                rule_ids.update(matcher.patterns)

        if "instead_of" in kwargs:
            matcher = self._overrides_matchers.get(kwargs["instead_of"])
            if matcher:
                if rule_ids:
                    # If both params are used then narrow down the rules.
                    rule_ids = {
                        k: v for k, v in matcher.patterns.items() if k in rule_ids
                    }
                else:
                    rule_ids.update(matcher.patterns)

        rules = [self._rules[id_] for id_ in rule_ids]

        if rules and kwargs.keys() <= {"to_return", "instead_of"}:
            return rules

        # Search other parameters as well

        getter = attrgetter(*kwargs.keys())

        def finder(rule: ApplyRule):
            attribs = getter(rule)
            if not isinstance(attribs, tuple):
                attribs = (attribs,)
            if attribs == tuple(kwargs.values()):
                return True
            return False

        results = []
        for rule in rules or self.get_rules():
            if finder(rule):
                results.append(rule)
        return results

    def search_overrides(self, **kwargs) -> List[ApplyRule]:
        """Deprecated, use :meth:`~.RulesRegistry.search` instead."""
        msg = "The 'search_overrides' method is deprecated. Use 'search' instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.search(**kwargs)

    def _match_url_for_page_object(
        self, url: Union[_Url, str], matcher: Optional[URLMatcher] = None
    ) -> Optional[Type[ItemPage]]:
        """Returns the page object to use based on the URL and URLMatcher."""
        if not url or matcher is None:
            return None

        rule_id = matcher.match(str(url))
        if rule_id is not None:
            return self._rules[rule_id].use

    def overrides_for(
        self, url: Union[_Url, str]
    ) -> Mapping[Type[ItemPage], Type[ItemPage]]:
        """Finds all of the page objects associated with the given URL and
        returns a Mapping where the 'key' represents the page object that is
        **overridden** by the page object in 'value'."""
        result: Dict[Type[ItemPage], Type[ItemPage]] = {}
        for replaced_page, matcher in self._overrides_matchers.items():
            if replaced_page is None:
                continue
            page = self._match_url_for_page_object(url, matcher)
            if page:
                result[replaced_page] = page
        return result

    def page_cls_for_item(
        self, url: Union[_Url, str], item_cls: Type
    ) -> Optional[Type]:
        """Return the page object class associated with the given URL that's able
        to produce the given ``item_cls``."""
        if item_cls is None:
            return None
        matcher = self._item_matchers.get(item_cls)
        return self._match_url_for_page_object(url, matcher)


def _walk_module(module: str) -> Iterable:
    """Return all modules from a module recursively.

    Note that this will import all the modules and submodules. It returns the
    provided module as well.
    """

    def onerror(err):
        raise err  # pragma: no cover

    spec = importlib.util.find_spec(module)
    if not spec:
        raise ImportError(f"Module {module} not found")
    mod = importlib.import_module(spec.name)
    yield mod
    if spec.submodule_search_locations:
        for info in pkgutil.walk_packages(
            spec.submodule_search_locations, f"{spec.name}.", onerror
        ):
            mod = importlib.import_module(info.name)
            yield mod


def consume_modules(*modules: str) -> None:
    """This recursively imports all packages/modules so that the ``@handle_urls``
    decorators are properly discovered and imported.

    Let's take a look at an example:

    .. code-block:: python

        # FILE: my_page_obj_project/load_rules.py

        from web_poet import default_registry, consume_modules

        consume_modules("other_external_pkg.po", "another_pkg.lib")
        rules = default_registry.get_rules()

    For this case, the :class:`~.ApplyRule` are coming from:

        - ``my_page_obj_project`` `(since it's the same module as the file above)`
        - ``other_external_pkg.po``
        - ``another_pkg.lib``
        - any other modules that was imported in the same process inside the
          packages/modules above.

    If the ``default_registry`` had other ``@handle_urls`` decorators outside of
    the packages/modules listed above, then the corresponding :class:`~.ApplyRule`
    won't be returned. Unless, they were recursively imported in some way similar
    to :func:`~.web_poet.rules.consume_modules`.
    """

    for module in modules:
        gen = _walk_module(module)

        # Inspired by itertools recipe: https://docs.python.org/3/library/itertools.html
        # Using a deque() results in a tiny bit performance improvement that list().
        deque(gen, maxlen=0)


OverrideRule = _create_deprecated_class("OverrideRule", ApplyRule, warn_once=False)
PageObjectRegistry = _create_deprecated_class(
    "PageObjectRegistry", RulesRegistry, warn_once=True
)



"""Test page object input validation scenarios."""

from typing import Optional

import attrs
import pytest

from web_poet import ItemPage, Returns, field, validates_input
from web_poet.exceptions import Retry, UseFallback


@attrs.define
class Item:
    a: str
    is_valid: bool = True


EXPECTED_ITEM = Item(a="a", is_valid=True)


class BasePage(ItemPage[Item]):
    @field
    def a(self):
        return "a"


# Valid input


class BaseValidInputPage(BasePage):
    def validate_input(self):
        pass


def test_valid_input_sync_to_item():
    class Page(BaseValidInputPage):
        def to_item(self):
            return Item(a=self.a)

    assert Page().to_item() == EXPECTED_ITEM


@pytest.mark.asyncio
async def test_valid_input_async_to_item():
    assert await BaseValidInputPage().to_item() == EXPECTED_ITEM


def test_valid_input_sync_field():
    assert BaseValidInputPage().a == "a"


@pytest.mark.asyncio
async def test_valid_input_async_field():
    class Page(BaseValidInputPage):
        @field
        async def a(self):
            return "a"

    assert await Page().a == "a"


# Retry


class BaseRetryPage(BasePage):
    def validate_input(self):
        raise Retry()


def test_retry_sync_to_item():
    class Page(BaseRetryPage):
        def to_item(self):
            return Item(a=self.a)

    page = Page()
    with pytest.raises(Retry):
        page.to_item()


@pytest.mark.asyncio
async def test_retry_async_to_item():
    page = BaseRetryPage()
    with pytest.raises(Retry):
        await page.to_item()


def test_retry_sync_field():
    page = BaseRetryPage()
    with pytest.raises(Retry):
        page.a


@pytest.mark.asyncio
async def test_retry_async_field():
    class Page(BaseRetryPage):
        @field
        async def a(self):
            return "a"

    page = Page()
    with pytest.raises(Retry):
        await page.a


# Use fallback


class BaseUseFallbackPage(BasePage):
    def validate_input(self):
        if self.a is None:
            raise UseFallback()

    @field
    def a(self):
        return None


def test_use_fallback_sync_to_item():
    class Page(BaseUseFallbackPage):
        def to_item(self):
            return Item(a=self.a)

    page = Page()
    with pytest.raises(UseFallback):
        page.to_item()


@pytest.mark.asyncio
async def test_use_fallback_async_to_item():
    page = BaseUseFallbackPage()
    with pytest.raises(UseFallback):
        await page.to_item()


def test_use_fallback_sync_field():
    page = BaseUseFallbackPage()
    with pytest.raises(UseFallback):
        page.a


@pytest.mark.asyncio
async def test_use_fallback_async_field():
    class Page(BaseUseFallbackPage):
        def validate_input(self):
            # Cannot use async self.a
            raise UseFallback()

        @field
        async def a(self):
            return "a"

    page = Page()
    with pytest.raises(UseFallback):
        await page.a


# Invalid input


INVALID_ITEM = Item(a="invalid", is_valid=False)


class BaseInvalidInputPage(ItemPage[Item]):
    def validate_input(self):
        return INVALID_ITEM

    @field
    def a(self):
        raise RuntimeError("This exception should never be raised")


def test_invalid_input_sync_to_item():
    class Page(BaseInvalidInputPage):
        @validates_input
        def to_item(self):
            return Item(a=self.a)

    assert Page().to_item() == INVALID_ITEM


@pytest.mark.asyncio
async def test_invalid_input_async_to_item():
    assert await BaseInvalidInputPage().to_item() == INVALID_ITEM


def test_invalid_input_sync_field():
    assert BaseInvalidInputPage().a == "invalid"


@pytest.mark.asyncio
async def test_invalid_input_async_field():
    class Page(BaseInvalidInputPage):
        @field
        async def a(self):
            raise RuntimeError("This exception should never be raised")

    assert await Page().a == "invalid"


# Unvalidated input


def test_unvalidated_input_sync_to_item():
    class Page(BasePage):
        def to_item(self):
            return Item(a=self.a)

    assert Page().to_item() == EXPECTED_ITEM


@pytest.mark.asyncio
async def test_unvalidated_input_async_to_item():
    assert await BasePage().to_item() == EXPECTED_ITEM


def test_unvalidated_input_sync_field():
    assert BasePage().a == "a"


@pytest.mark.asyncio
async def test_unvalidated_input_async_field():
    class Page(BasePage):
        @field
        async def a(self):
            return "a"

    assert await Page().a == "a"


# Caching


class BaseCachingPage(BasePage):
    _raise = False

    def validate_input(self):
        if self._raise:
            raise UseFallback()
        self._raise = True


def test_invalid_input_sync_to_item_caching():
    class Page(BaseCachingPage):
        def to_item(self):
            return Item(a=self.a)

    page = Page()
    page.to_item()
    page.to_item()


@pytest.mark.asyncio
async def test_invalid_input_async_to_item_caching():
    page = BaseCachingPage()
    await page.to_item()
    await page.to_item()


def test_invalid_input_sync_field_caching():
    page = BaseCachingPage()
    page.a
    page.a


@pytest.mark.asyncio
async def test_invalid_input_async_field_caching():
    class Page(BaseCachingPage):
        @field
        async def a(self):
            return "a"

    page = Page()
    await page.a
    await page.a


@pytest.mark.asyncio
async def test_invalid_input_cross_api_caching():
    @attrs.define
    class _Item(Item):
        b: Optional[str] = None

    class Page(BaseCachingPage, Returns[_Item]):
        @field
        async def b(self):
            return "b"

    page = Page()
    page.a
    await page.b
    await page.to_item()


# Recursion


@pytest.mark.asyncio
async def test_recursion():
    """Make sure that using fields within the validate_input method does not
    result in a recursive call to the validate_input method."""

    class Page(BasePage):
        _raise = False

        def validate_input(self):
            if self._raise:
                raise UseFallback()
            self._raise = True
            assert self.a == "a"

    page = Page()
    assert page.a == "a"



import warnings

import attrs
import pytest
from url_matcher import Patterns

from tests.po_lib import (
    POTopLevel1,
    POTopLevel2,
    POTopLevelOverriden1,
    POTopLevelOverriden2,
)
from tests.po_lib.a_module import POModule, POModuleOverriden
from tests.po_lib.nested_package import PONestedPkg, PONestedPkgOverriden
from tests.po_lib.nested_package.a_nested_module import (
    PONestedModule,
    PONestedModuleOverriden,
)
from tests.po_lib_sub import POLibSub
from tests.po_lib_to_return import (
    CustomProductPage,
    CustomProductPageDataTypeOnly,
    CustomProductPageNoReturns,
    ImprovedProductPage,
    LessProductPage,
    MoreProductPage,
    Product,
    ProductFewerFields,
    ProductMoreFields,
    ProductPage,
    ProductSeparate,
    ProductSimilar,
    SeparateProductPage,
    SimilarProductPage,
    SomePage,
)
from web_poet import (
    ApplyRule,
    OverrideRule,
    RulesRegistry,
    consume_modules,
    default_registry,
    handle_urls,
)
from web_poet.page_inputs.url import RequestUrl, ResponseUrl

POS = {
    CustomProductPage,
    CustomProductPageNoReturns,
    CustomProductPageDataTypeOnly,
    ImprovedProductPage,
    LessProductPage,
    MoreProductPage,
    POTopLevel1,
    POTopLevel2,
    POModule,
    PONestedPkg,
    PONestedModule,
    ProductPage,
    SeparateProductPage,
    SimilarProductPage,
    SomePage,
}


def test_apply_rule_uniqueness() -> None:
    """The same instance of an ApplyRule with the same attribute values should
    have the same hash identity.
    """

    patterns = Patterns(include=["example.com"], exclude=["example.com/blog"])
    patterns_b = Patterns(include=["example.com/b"])

    rule1 = ApplyRule(
        for_patterns=patterns,
        use=POTopLevel1,
        instead_of=POTopLevelOverriden1,
        meta={"key_1": 1},
    )
    rule2 = ApplyRule(
        for_patterns=patterns,
        use=POTopLevel1,
        instead_of=POTopLevelOverriden1,
        meta={"key_2": 2},
    )
    # The ``meta`` parameter is ignored in the hash.
    assert hash(rule1) == hash(rule2)

    params = [
        {
            "for_patterns": patterns,
            "use": POTopLevel1,
            "instead_of": POTopLevelOverriden1,
            "to_return": Product,
        },
        {
            "for_patterns": patterns_b,
            "use": POTopLevel2,
            "instead_of": POTopLevelOverriden2,
            "to_return": ProductSimilar,
        },
    ]

    for change in params[0].keys():
        # Changing any one of the params should result in a hash mismatch
        rule1 = ApplyRule(**params[0])
        kwargs = params[0].copy()
        kwargs.update({change: params[1][change]})
        rule2 = ApplyRule(**kwargs)
        assert hash(rule1) != hash(rule2)


def test_apply_rule_immutability() -> None:
    patterns = Patterns(include=["example.com"], exclude=["example.com/blog"])

    rule = ApplyRule(
        for_patterns=patterns,
        use=POTopLevel1,
        instead_of=POTopLevelOverriden1,
    )

    with pytest.raises(attrs.exceptions.FrozenInstanceError):
        rule.for_patterns = Patterns(include=["example.com/"])  # type: ignore[misc]

    with pytest.raises(attrs.exceptions.FrozenInstanceError):
        rule.use = POTopLevel2  # type: ignore[misc]

    with pytest.raises(attrs.exceptions.FrozenInstanceError):
        rule.instead_of = POTopLevelOverriden2  # type: ignore[misc]


def test_apply_rule_converter_on_pattern() -> None:
    # passing strings should auto-converter into Patterns
    rule = ApplyRule("example.com", use=POTopLevel1, instead_of=POTopLevelOverriden2)
    assert rule.for_patterns == Patterns(
        include=("example.com",), exclude=(), priority=500
    )

    # Passing Patterns should still work
    rule = ApplyRule(
        for_patterns=Patterns(["example.com"]),
        use=POTopLevel1,
        instead_of=POTopLevelOverriden2,
    )
    assert rule.for_patterns == Patterns(
        include=("example.com",), exclude=(), priority=500
    )


def test_apply_rule_kwargs_only() -> None:
    params = {
        "use": POTopLevel1,
        "instead_of": POTopLevelOverriden2,
        "to_return": Product,
        "meta": {"key_2": 2},
    }
    remove = set()

    for param_name in params:
        remove.add(param_name)
        with pytest.raises(TypeError):
            ApplyRule(
                "example.com",
                *[params[r] for r in remove],
                **{k: v for k, v in params.items() if k not in remove},  # type: ignore[arg-type]  # noqa: B038
            )


def test_list_page_objects_all() -> None:
    rules = default_registry.get_rules()
    page_objects = {po.use for po in rules}

    # Note that the 'tests_extra.po_lib_sub_not_imported.POLibSubNotImported'
    # Page Object is not included here since it was never imported anywhere in
    # our test package. It would only be included if we run any of the following
    # below. (Note that they should run before `get_rules` is called.)
    #   - from tests_extra import po_lib_sub_not_imported
    #   - import tests_extra.po_lib_sub_not_imported
    #   - web_poet.consume_modules("tests_extra")
    # Merely having `import tests_extra` won't work since the subpackages and
    # modules needs to be traversed and imported as well.
    assert all("po_lib_sub_not_imported" not in po.__module__ for po in page_objects)

    # Ensure that ALL Override Rules are returned as long as the given
    # registry's @handle_urls decorator was used.
    assert page_objects == POS.union({POLibSub})

    for rule in rules:
        # We're ignoring the types below since mypy expects ``Type[ItemPage]``
        # which doesn't contain the ``expected_*`` fields in our tests.

        # Special case since this PO has 2 ``@handle_urls`` decorators.
        # See ``test_multiple_handle_urls()`` test case below.
        if rule.use == POTopLevel1:
            continue

        assert rule.instead_of == rule.use.expected_instead_of, rule.use  # type: ignore[attr-defined]
        assert rule.for_patterns == rule.use.expected_patterns, rule.use  # type: ignore[attr-defined]
        assert rule.to_return == rule.use.expected_to_return, rule.use  # type: ignore[attr-defined]
        assert rule.meta == rule.use.expected_meta, rule.use  # type: ignore[attr-defined]


def test_multiple_handle_urls_annotations() -> None:
    """Using multiple ``@handle_urls`` annotations on a single Page Object
    should work.
    """
    rules = default_registry.search(use=POTopLevel1)
    assert len(rules) == 2

    for i, rule in enumerate(rules):
        assert rule.instead_of == rule.use.expected_instead_of[i], rule.use  # type: ignore[attr-defined]
        assert rule.for_patterns == rule.use.expected_patterns[i], rule.use  # type: ignore[attr-defined]
        assert rule.to_return == rule.use.expected_to_return[i], rule.use  # type: ignore[attr-defined]
        assert rule.meta == rule.use.expected_meta[i], rule.use  # type: ignore[attr-defined]


def test_registry_get_overrides_deprecation() -> None:
    msg = "The 'get_overrides' method is deprecated. Use 'get_rules' instead."
    with pytest.warns(DeprecationWarning, match=msg):
        rules = default_registry.get_overrides()

    # It should still work as usual
    assert len(rules) == len(default_registry.get_rules())

    # but the rules from ``.get_overrides()`` should return ``ApplyRule`` and
    # not the old ``OverrideRule``.
    assert all(r for r in rules if isinstance(r, ApplyRule))


def test_consume_module_not_existing() -> None:
    with pytest.raises(ImportError):
        consume_modules("this_does_not_exist")


def test_list_page_objects_all_consume() -> None:
    """A test similar to the one above but calls ``consume_modules()`` to properly
    load the ``@handle_urls`` decorators from other modules/packages.
    """
    consume_modules("tests_extra")
    rules = default_registry.get_rules()
    page_objects = {po.use for po in rules}
    assert any("po_lib_sub_not_imported" in po.__module__ for po in page_objects)


def test_registry_search() -> None:
    # param: use
    rules = default_registry.search(use=POTopLevel2)
    assert len(rules) == 1
    assert rules[0].use == POTopLevel2

    # param: instead_of
    rules = default_registry.search(instead_of=POTopLevelOverriden2)
    assert len(rules) == 1
    assert rules[0].instead_of == POTopLevelOverriden2

    rules = default_registry.search(instead_of=None)
    for rule in rules:
        assert rule.instead_of is None

    # param: to_return
    rules = default_registry.search(to_return=Product)
    assert rules == [
        ApplyRule("example.com", use=ProductPage, to_return=Product),
        ApplyRule(
            "example.com",
            use=ImprovedProductPage,
            instead_of=ProductPage,
            to_return=Product,
        ),
        ApplyRule(
            "example.com",
            # mypy complains here since it's expecting a container class when
            # declared, i.e, ``ItemPage[SomeItem]``
            use=CustomProductPageDataTypeOnly,  # type: ignore[arg-type]
            to_return=Product,
        ),
    ]

    rules = default_registry.search(to_return=None)
    for rule in rules:
        assert rule.to_return is None

    # params: to_return and use
    rules = default_registry.search(to_return=Product, use=ImprovedProductPage)
    assert len(rules) == 1
    assert rules[0].to_return == Product
    assert rules[0].use == ImprovedProductPage

    # params: to_return and instead_of
    rules = default_registry.search(to_return=Product, instead_of=None)
    assert len(rules) == 2
    assert rules[0].to_return == Product
    assert rules[0].instead_of is None
    assert rules[1].to_return == Product
    assert rules[1].instead_of is None

    rules = default_registry.search(to_return=None, instead_of=ProductPage)
    for rule in rules:
        assert rule.to_return is None
        assert rule.instead_of is None

    rules = default_registry.search(to_return=None, instead_of=None)
    assert len(rules) == 1
    assert rules[0].to_return is None
    assert rules[0].instead_of is None

    # Such rules doesn't exist
    rules = default_registry.search(use=POModuleOverriden)
    assert len(rules) == 0


def test_registry_search_overrides_deprecation() -> None:
    msg = "The 'search_overrides' method is deprecated. Use 'search' instead."
    with pytest.warns(DeprecationWarning, match=msg):
        rules = default_registry.search_overrides(use=POTopLevel2)

    # It should still work as usual
    assert len(rules) == 1
    assert rules[0].use == POTopLevel2

    # The rules from ``.get_overrides()`` should return ``ApplyRule`` and
    # not the old ``OverrideRule``.
    assert isinstance(rules[0], ApplyRule)


def test_init_rules() -> None:
    rules = (
        ApplyRule(
            for_patterns=Patterns(include=["example.com"]),
            use=POTopLevel1,
            instead_of=POTopLevelOverriden2,
        ),
    )

    registry = RulesRegistry(rules=rules)

    # Any type of iterable input should convert it to a list.
    assert registry.get_rules() == list(rules)
    assert default_registry.get_rules() != rules


def test_add_rule() -> None:
    registry = RulesRegistry()

    # Basic case of adding a rule
    rule_1 = ApplyRule(
        for_patterns=Patterns(include=["example.com"]),
        use=POTopLevel1,
        instead_of=POTopLevelOverriden1,
        to_return=Product,
    )
    registry.add_rule(rule_1)
    assert registry.get_rules() == [rule_1]

    # Adding a second rule should not emit a warning as long as both the URL
    # pattern and `.to_return` value is not the same.
    rule_2 = ApplyRule(
        for_patterns=Patterns(include=["example.com"]),
        use=POTopLevel1,
        instead_of=POTopLevelOverriden2,
        to_return=ProductSimilar,
    )
    with warnings.catch_warnings(record=True) as warnings_emitted:
        registry.add_rule(rule_2)
    assert not warnings_emitted
    assert registry.get_rules() == [rule_1, rule_2]

    # Warnings should be raised for this case since it's the same URL pattern
    # and `.to_return` value from one of the past rules.
    rule_3 = ApplyRule(
        for_patterns=Patterns(include=["example.com"]),
        use=POTopLevel1,
        instead_of=POTopLevelOverriden2,
        to_return=Product,
    )
    with pytest.warns(UserWarning, match="conflicting rules"):
        registry.add_rule(rule_3)

    assert registry.get_rules() == [rule_1, rule_2, rule_3]


def test_overrides_for() -> None:
    for cls in [str, RequestUrl, ResponseUrl]:
        assert default_registry.overrides_for(cls("https://example.com")) == {
            POTopLevelOverriden1: POTopLevel1,
            POTopLevelOverriden2: POTopLevel2,
            POModuleOverriden: POModule,
            PONestedPkgOverriden: PONestedPkg,
            PONestedModuleOverriden: PONestedModule,
            ProductPage: CustomProductPageNoReturns,
        }

        assert default_registry.overrides_for(cls("https://example.org")) == {
            PONestedModuleOverriden: PONestedModule,
            PONestedPkgOverriden: PONestedPkg,
        }


def test_page_cls_for_item() -> None:
    # This is not associated with any rule.
    class FakeItem:
        pass

    method = default_registry.page_cls_for_item

    for cls in [str, RequestUrl, ResponseUrl]:
        url = cls("https://example.com")
        assert method(url, ProductSimilar) == CustomProductPageNoReturns
        assert method(url, Product) == CustomProductPageDataTypeOnly
        assert method(url, ProductSeparate) == SeparateProductPage
        assert method(url, ProductFewerFields) == LessProductPage
        assert method(url, ProductMoreFields) == MoreProductPage

        # Type is ignored since item_cls shouldn't be None
        assert method(url, None) is None  # type: ignore[arg-type]

        # When there's no rule specifying to return this FakeItem
        assert method(url, FakeItem) is None

        # When the URL itself doesn't have any ``to_return`` in any of its rules
        assert method(cls("https://example.org"), FakeItem) is None


def test_from_override_rules_deprecation_using_ApplyRule() -> None:
    rules = [
        ApplyRule(
            for_patterns=Patterns(include=["example.com"]),
            use=POTopLevel1,
            instead_of=POTopLevelOverriden2,
        )
    ]

    msg = "The 'from_override_rules' method is deprecated."
    with pytest.warns(DeprecationWarning, match=msg):
        registry = RulesRegistry.from_override_rules(rules)

    assert registry.get_rules() == rules
    assert default_registry.get_rules() != rules


def test_from_override_rules_deprecation_using_OverrideRule() -> None:
    rules = [
        OverrideRule(
            for_patterns=Patterns(include=["example.com"]),
            use=POTopLevel1,
            instead_of=POTopLevelOverriden2,
        )
    ]

    msg = "The 'from_override_rules' method is deprecated."
    with pytest.warns(DeprecationWarning, match=msg):
        registry = RulesRegistry.from_override_rules(rules)

    assert registry.get_rules() == rules
    assert default_registry.get_rules() != rules


def test_handle_urls_deprecation() -> None:
    before_count = len(default_registry.get_rules())

    msg = (
        "The 'overrides' parameter in @handle_urls is deprecated. Use the "
        "'instead_of' parameter."
    )
    with pytest.warns(DeprecationWarning, match=msg):

        @handle_urls("example.com", overrides=CustomProductPage)
        class PageWithDeprecatedOverrides: ...

    # Despite the deprecation, it should still properly add the rule in the
    # registry.
    after_count = len(default_registry.get_rules())
    assert after_count == before_count + 1

    # The added rule should have its deprecated 'overrides' parameter converted
    # into the new 'instead_of' parameter.
    rules = default_registry.search(
        instead_of=CustomProductPage, use=PageWithDeprecatedOverrides
    )
    assert rules == [
        ApplyRule(
            "example.com",
            instead_of=CustomProductPage,
            # mypy complains here since it's expecting a container class when
            # declared, i.e, ``ItemPage[SomeItem]``
            use=PageWithDeprecatedOverrides,  # type: ignore[arg-type]
        )
    ]


def test_override_rule_deprecation() -> None:
    msg = (
        "web_poet.rules.OverrideRule is deprecated, "
        "instantiate web_poet.rules.ApplyRule instead."
    )
    with pytest.warns(DeprecationWarning, match=msg):
        OverrideRule(for_patterns=None, use=None)
